@@ -5,11 +5,10 @@ use warnings;
 
 use base qw| HTML::Mason::CGIHandler |;
 
-#
-# Parent does funny things with eval before we can.
-#
-# Delegate to H::M::R instead.
-#
+### CGIHandler evals stuff before we can check on the result code,
+### which can make it hard or impossible to trap specific errors.
+###
+### Avoid this by delegating to H~::M~::Request's exec method instead.
 sub exec {
   my $self = shift;
 
@@ -18,13 +17,47 @@ sub exec {
 
 package Devel::Pillbug;
 
-our $VERSION = 0.005;
+our $VERSION = 0.006;
 
 use strict;
 use warnings;
 
 use File::HomeDir;
-use File::Type;
+use Media::Type::Simple;
+
+### Media::Type::Simple's internal use of a cached filehandle makes
+### it not usable when forking, and its internal use of private globals
+### makes it hard to subclass.
+###
+### Sadly, Media::Type::Simple is currently the best thing going on CPAN
+### in terms of guessing MIME types from file extensions, so... until
+### its author applies a fix for this very common problem (or something
+### better comes along), I am adapting Jos Boumans's workaround from
+### rt.cpan.org #46474:
+do {
+  no strict "refs";
+
+  *{"Media::Type::Simple::__new"} = sub {
+    my $class = shift;
+    my $self = { types => {}, extens => {}, };
+
+    bless $self, $class;
+
+    if (@_) {
+      my $fh = shift;
+      return $self->add_types_from_file($fh);
+    } else {
+      my $offset = tell Media::Type::Simple::DATA;
+
+      $Media::Type::Simple::Default =
+        $self->add_types_from_file( \*Media::Type::Simple::DATA );
+
+      seek Media::Type::Simple::DATA, $offset, 0;
+
+      return clone $Media::Type::Simple::Default;
+    }
+    }
+};
 
 use base qw| HTTP::Server::Simple::Mason |;
 
@@ -73,6 +106,83 @@ sub handler_class {
   return $handlerClass;
 }
 
+#
+#
+#
+sub pretty_html_header {
+  my $self   = shift;
+  my $header = shift;
+
+  if ($header) {
+    $self->{_html_header} = $header;
+
+    return;
+  }
+
+  if ( defined $self->{_html_header} ) {
+    print $self->{_html_header};
+
+    return;
+  }
+
+  print "<html>\n";
+  print "<head>\n";
+  print "<style>\n";
+  print "  body {\n";
+  print "    background: #fff; color: #333;\n";
+  print "  }\n";
+  print "  body, td {\n";
+  print "    font-family: verdana, sans-serif;\n";
+  print "    font-size: 11px;\n";
+  print "  }\n";
+  print "  td {\n";
+  print "    border-bottom: 1px dotted #999;\n";
+  print "  }\n";
+  print "  h1 {\n";
+  print "    color: #999;\n";
+  print "  }\n";
+  print "  pre {\n";
+  print "    white-space: pre-wrap;\n";
+  print "    background: #ccc; color: #000;\n";
+  print "    margin: 6px; padding: 6px;\n";
+  print "  }\n";
+  print "</style>\n";
+  print "</head>\n";
+  print "<body>\n";
+}
+
+#
+#
+#
+sub pretty_html_footer {
+  my $self   = shift;
+  my $footer = shift;
+
+  if ($footer) {
+    $self->{_html_footer} = $footer;
+
+    return;
+  }
+
+  if ( $self->{_html_footer} ) {
+    print $self->{_html_footer};
+
+    return;
+  }
+
+  my @time = localtime();
+  my $time = sprintf(
+    '%i-%02d-%02d %02d:%02d:%02d %s',
+    $time[5] + 1900,
+    $time[4] + 1,
+    $time[3], $time[2], $time[1], $time[0], POSIX::strftime( '%Z', @time )
+  );
+
+  print "<p>$time</p>\n";
+  print "</body>\n";
+  print "</html>\n";
+}
+
 #
 #
 #
@@ -159,31 +269,48 @@ sub _handle_mason_request {
   my $path = shift;
 
   my $r = HTML::Mason::FakeApache->new( cgi => $cgi );
-
-  my $m = $self->mason_handler;
-
-  my $comp = $m->interp->make_component( comp_file => $path );
-
   my $buffer;
 
-  my $req = $m->interp->make_request(
-    comp        => $comp,
-    args        => [ $cgi->Vars ],
-    cgi_request => $r,
-    out_method  => \$buffer,
-  );
+  ###
+  ### Brutal and tempoorary workaround for undef warnings caused
+  ### by anonymous components calling other components.
+  ###
+  ### https://rt.cpan.org/Public/Bug/Display.html?id=55159
+  do {
+    no strict "refs";
+    no warnings "redefine";
 
-  $r->{http_header_sent} = 1;
+    *{"HTML::Mason::Component::dir_path"} = sub { return "" };
+  };
 
-  $m->interp->set_global( '$r', $r );
+  eval {
+    my $m = $self->mason_handler;
 
-  HTML::Mason::Request::exec($req);
+    my $comp = $m->interp->make_component( comp_file => $path );
+
+    my $req = $m->interp->make_request(
+      comp         => $comp,
+      args         => [ $cgi->Vars ],
+      cgi_request  => $r,
+      out_method   => \$buffer,
+      error_mode   => "fatal",
+      error_format => "text",
+    );
+
+    $r->{http_header_sent} = 1;
+
+    $m->interp->set_global( '$r', $r );
+
+    $req->exec;
+  };
 
   #
   #
   #
   if ( $@ && ( !$r->status || ( $r->status !~ /^302/ ) ) ) {
     $r->status("500 Internal Server Error");
+
+    return $self->_handle_error( $r, $@ );
   } elsif ( !$r->status ) {
     $r->status("200 OK");
   }
@@ -209,19 +336,69 @@ sub _handle_directory_request {
   print "HTTP/1.0 200 OK\r\n";
   print "Content-Type: text/html\r\n";
   print "\r\n";
-  print "<h1>Index of $compPath</h1>\r\n";
-  print "<ul>\r\n";
+
+  $self->pretty_html_header();
+
+  print "<h1>Index of $compPath</h1>\n";
+
+  print "<table width=\"100%\" cellspacing=\"0\" cellpadding=\"2\">\n";
+  print "  <tr>\n";
+  print "    <td>Name</td>\n";
+  print "    <td>Type</td>\n";
+  print "    <td>Last Modified</td>\n";
+  print "    <td>Size</td>\n";
+  print "  </tr>\n";
 
   my %conf = $self->mason_config;
 
-  for ( <$fsPath/*> ) {
+  my @files;
+
+  if ( $compPath ne "/" ) { push @files, ".." }
+
+  for (<$fsPath/*>) { push @files, $_ }
+
+  for (@files) {
     my $path = $_;
+
+    my @stat = stat($path);
+
+    my $type;
+    my $size;
+
+    if ( -d $path ) {
+      $path .= '/';
+      $type = "directory";
+      $size = "-";
+    } else {
+      my $ext = $path;
+      $ext =~ s/.*\.//;
+      my $o = Media::Type::Simple->__new();
+      eval { $type = $o->type_from_ext($ext); };
+      $type ||= "application/octet-stream";
+      $size = $stat[7];
+    }
+
     $path =~ s/^$conf{comp_root}$compPath\///;
 
-    print "<li> <a href=\"$path\">$path</a></li>\r\n";
+    my @time = localtime( $stat[9] );
+    my $time = sprintf(
+      '%i-%02d-%02d %02d:%02d:%02d %s',
+      $time[5] + 1900,
+      $time[4] + 1,
+      $time[3], $time[2], $time[1], $time[0], POSIX::strftime( '%Z', @time )
+    );
+
+    print "  <tr>\n";
+    print "    <td><a href=\"$path\">$path</a></td>\n";
+    print "    <td>$type</td>\n";
+    print "    <td>$time</td>\n";
+    print "    <td>${ size }</td>\n";
+    print "  </tr>\n";
   }
 
-  print "</ul>\r\n";
+  print "</table>\n";
+
+  $self->pretty_html_footer();
 }
 
 sub _handle_document_request {
@@ -231,8 +408,12 @@ sub _handle_document_request {
   my $fsPath   = shift;
   my $compPath = shift;
 
-  my $ft   = File::Type->new();
-  my $type = $ft->mime_type($fsPath);
+  my $ext = $fsPath;
+  $ext =~ s/.*\.//;
+  my $o = Media::Type::Simple->__new();
+  my $type;
+  eval { $type = $o->type_from_ext($ext); };
+  $type ||= "application/octet-stream";
 
   my @out;
 
@@ -242,8 +423,8 @@ sub _handle_document_request {
     close(IN);
   };
 
-  if ( $@ ) {
-    return $self->_handle_error($r, $@);
+  if ($@) {
+    return $self->_handle_error( $r, $@ );
   }
 
   print "HTTP/1.0 200 OK\r\n";
@@ -263,8 +444,13 @@ sub _handle_notfound_request {
   print "HTTP/1.0 404 Not Found\r\n";
   print "Content-Type: text/html\r\n";
   print "\r\n";
-  print "<h1>Not Found</h1>\r\n";
-  print "<p>The requested URL $compPath was not found on this server.\r\n";
+
+  $self->pretty_html_header();
+
+  print "<h1>Not Found</h1>\n";
+  print "<p>The requested URL $compPath was not found on this server.\n";
+
+  $self->pretty_html_footer();
 }
 
 sub _handle_error {
@@ -273,18 +459,25 @@ sub _handle_error {
 
   my $err = shift;
 
-  $err =~ s/at \S+ line \d+.*//;
+  # $err =~ s/at \S+ line \d+.*//;
+
+  $err = HTML::Entities::encode_entities($err);
 
   print "HTTP/1.0 500 Internal Server Error\r\n";
   print "Content-type: text/html\r\n";
   print "\r\n";
-  print "<h1>Internal Server Error</h1>\r\n";
-  print "<p>The server could not complete your request. The error was:</p>\r\n";
-  print "<p>$err</p>\r\n";
+
+  $self->pretty_html_header();
+
+  print "<h1>Internal Server Error</h1>\n";
+  print "<p>The server could not complete your request. The error was:</p>\n";
+  print "<pre>$err</pre>\n";
+
+  $self->pretty_html_footer();
 }
 
 sub _handle_directory_redirect {
-  my $self = shift;
+  my $self     = shift;
   my $compPath = shift;
 
   my $url = sprintf 'http://%s:%s%s/', $self->host, $self->port, $compPath;
@@ -292,12 +485,18 @@ sub _handle_directory_redirect {
   print "HTTP/1.0 302 Moved\r\n";
   print "Location: $url\r\n";
   print "\r\n";
-  print "<h1>Moved</h1>\r\n";
-  print "<p>The document is available <a href=\"$url\">here</a>.</p>\r\n";
+
+  $self->pretty_html_header();
+
+  print "<h1>Moved</h1>\n";
+  print "<p>The document is available <a href=\"$url\">here</a>.</p>\n";
+
+  $self->pretty_html_footer();
 }
 
 #
-# Adapted from H::S::S::Mason
+# Override HTTP::Server::Simple::Mason to also deal with document requests,
+# directory listings, and 404s
 #
 sub handle_request {
   my $self = shift;
@@ -332,18 +531,25 @@ sub handle_request {
     }
   }
 
-  if ( $compPath =~ /$ext$/ && $m->interp->comp_exists($compPath) ) {
-    $self->_handle_mason_request( $r, $fsPath, $compPath );
+  eval {
+    if ( $compPath =~ /$ext$/ && $m->interp->comp_exists($compPath) )
+    {
+      $self->_handle_mason_request( $r, $fsPath, $compPath );
 
-  } elsif ( $self->allow_index && -d $fsPath ) {
-    $self->_handle_directory_request( $r, $fsPath, $compPath );
+    } elsif ( $self->allow_index && -d $fsPath ) {
+      $self->_handle_directory_request( $r, $fsPath, $compPath );
 
-  } elsif ( !-d $fsPath && -e $fsPath ) {
-    $self->_handle_document_request( $r, $fsPath, $compPath );
+    } elsif ( !-d $fsPath && -e $fsPath ) {
+      $self->_handle_document_request( $r, $fsPath, $compPath );
 
-  } else {
-    $self->_handle_notfound_request( $r, $fsPath, $compPath );
+    } else {
+      $self->_handle_notfound_request( $r, $fsPath, $compPath );
 
+    }
+  };
+
+  if ($@) {
+    warn $@;
   }
 }
 
@@ -354,7 +560,7 @@ __END__
 
 =head1 NAME
 
-Devel::Pillbug - Tiny HTML::Mason server
+Devel::Pillbug - Stand-alone HTML::Mason-enabled server
 
 =head1 SYNOPSIS
 
@@ -388,17 +594,22 @@ Do it in Perl:
   #
   # $server->docroot("/tmp/foo");
 
+  #
+  # See docs or "pillbug -h" for further options
+  #
+
   $server->run;
 
 =head1 DESCRIPTION
 
-Devel::Pillbug is a tiny embedded L<HTML::Mason> server, based on
+Devel::Pillbug is a stand-alone L<HTML::Mason> server, extending
 L<HTTP::Server::Simple::Mason>. It is designed for zero configuration
 and easy install from CPAN.
 
 The "public_html" or "Sites" directory of the user who launched the
-process will be used for the document root. Files ending in "html"
-are treated as Mason components.
+process will be used for the default document root. Files ending
+in "html" are treated as Mason components. These and other behaviors
+may be overridden as needed.
 
 =head1 METHODS
 
@@ -409,7 +620,7 @@ inherited methods.
 
 =over 4
 
-=item * $class->net_server($newServerType);
+=item * $class->net_server($newServerType)
 
 Returns the currently active L<Net::Server> subclass.
 
@@ -418,7 +629,7 @@ is supplied as an argument.
 
 Default value is L<Net::Server::PreFork>.
 
-=item * $class->handler_class($newHandlerClass);
+=item * $class->handler_class($newHandlerClass)
 
 Returns the currently active L<HTML::Mason::Request> subclass.
 
@@ -433,25 +644,25 @@ Default value is L<Devel::Pillbug::MasonHandler>.
 
 =over 4
 
-=item * $self->docroot($docroot);
+=item * $self->docroot($docroot)
 
 Returns the currently active docroot.
 
 The server will set its docroot to the received absolute path, if
 supplied as an argument.
 
-=item * $self->index_name($name);
+=item * $self->index_name($name)
 
 Returns currently used index name, without extension (default is
 "index").
 
 Sets this to the received name, if supplied as an argument.
 
-=item * $self->comp_ext($extension);
+=item * $self->comp_ext($extension)
 
 Sets the file extension used for Mason components (default is "html")
 
-=item * $self->allow_index($bool);
+=item * $self->allow_index($bool)
 
 Returns the current allowed state for directory indexes.
 
@@ -459,6 +670,21 @@ Sets this to the received state, if supplied as an argument.
 
 0 = Off, 1 = On
 
+=item * $self->pretty_html_header($fragment)
+
+Prints the HTML fragment used for everything up to and including
+the "<body>" tag of internally-generated documents (errors and
+directory listings).
+
+Sets the fragment to the received string, if supplied as an argument.
+
+=item * $self->pretty_html_footer($fragment)
+
+Prints the HTML fragment used for everything below and including
+the "</body>" tag of internally-generated documents.
+
+Sets the fragment to the received string, if supplied as an argument.
+
 =back
 
 =head1 CONFIGURATION AND ENVIRONMENT
@@ -466,9 +692,20 @@ Sets this to the received state, if supplied as an argument.
 The document root must exist and be readable, and Devel::Pillbug
 must be able to bind to its listen port (default 8080).
 
+=head1 BUGS
+
+Absolutely...
+
+Currently, several brutish hacks are employed to work around minor
+issues in modules which Pillbug needs. These hacks will need to go
+away and/or be revisited over time.
+
+Please use the CPAN RT system or contact me if you find something
+which isn't working as advertised.
+
 =head1 VERSION
 
-This document is for version .005 of Devel::Pillbug.
+This document is for version .006 of Devel::Pillbug.
 
 =head1 AUTHOR
 
@@ -484,7 +721,7 @@ http://dev.perl.org/licenses/
 
 =head1 SEE ALSO
 
-L<File::HomeDir>, L<File::Type>, L<Net::Server::PreFork>.
+L<File::HomeDir>, L<Media::Type::Simple>, L<Net::Server::PreFork>.
 
 This module extends L<HTTP::Server::Simple::Mason>.
 
